#include <arc/job_xrsl.h>
#include <arc/stringconv.h>

static void add_attribute(const std::string& name,const std::string& value,Xrsl& xrsl) throw(XrslError) {
	if(value.length()) {
		xrsl.AddRelation(XrslRelation(name,operator_eq,value));
	};
}

static void add_attribute(const std::string& name,long value,Xrsl& xrsl) throw(XrslError) {
	if(value>=0) {
		xrsl.AddRelation(XrslRelation(name,operator_eq,tostring<long>(value)));
	};
}

static void add_attribute(const std::string& name,const std::list<std::string>& values,Xrsl& xrsl) throw(XrslError) {
	if(values.size()) {
		xrsl.AddRelation(XrslRelation(name,operator_eq,values));
	};
}

static void add_attributes(const std::string& name,const std::list<std::string>& values,Xrsl& xrsl) throw(XrslError) {
	for(std::list<std::string>::const_iterator i = values.begin();
	                                           i!=values.end();++i) {
		add_attribute(name,*i,xrsl);
	};
}

static void get_attribute(const std::string& name,std::string& value,Xrsl& xrsl) {
    try {
        value=(xrsl.GetRelation(name)).GetSingleValue();
    } catch (XrslError) { };
}

static void get_attribute(const std::string& name,int& value,Xrsl& xrsl) {
    try {
        value=stringto<int>((xrsl.GetRelation(name)).GetSingleValue());
    } catch (XrslError) {
	} catch (StringConvError) { };
}

static void get_attribute(const std::string& name,long& value,Xrsl& xrsl) {
    try {
        value=stringto<long>((xrsl.GetRelation(name)).GetSingleValue());
    } catch (XrslError) {
    } catch (StringConvError) { };
}

static std::string notification(const JobRequest::Notification v) {
	return v.flags+" "+v.email;
}

static std::string notification(const std::list<JobRequest::Notification> l) {
	std::string n;
	for(std::list<JobRequest::Notification>::const_iterator i = l.begin();
	                                                       i!=l.end();++i) {
		if(n.length()) n+=" ";
		n+=notification(*i);
	};
	return n;
}

static std::list<JobRequest::Notification> notification(const std::string& s) {
	std::list<JobRequest::Notification> n;
	std::size_t start = 0;
	std::size_t end = 0;
	std::string flags("");
	for(;;) {
		start=s.find_first_not_of(' ',end);
		if(start == std::string::npos) break;
		end=s.find(' ',start);
		if(end == std::string::npos) end=s.length();
		std::size_t p = s.find('@',start);
		if((p == std::string::npos) || (p >= end)) {
			flags=s.substr(start,end-start);
			continue;
		};
		if(flags.length() == 0) continue;
		n.push_back(JobRequest::Notification(flags,s.substr(start,end-start)));
	};
	return n;
}


bool JobRequestXRSL::set(const char* s) throw(JobRequestError) {
	reset();
	if(xrsl_) delete xrsl_;
	try {
		xrsl_=new Xrsl(std::string(s));
	} catch (XrslError e) {
		xrsl_=NULL;
		return false;
	};
    return set(*xrsl_);
}

bool JobRequestXRSL::set(Xrsl& xrsl) throw(JobRequestError) {
	// Fill internal fields
	get_attribute("jobname",job_name,xrsl);
	get_attribute("acl",acl,xrsl);
	get_attribute("stdin",stdin,xrsl);
	get_attribute("stdout",stdout,xrsl);
	get_attribute("stderr",stderr,xrsl);
	get_attribute("queue",queue,xrsl);
	get_attribute("gmlog",gmlog,xrsl);
	get_attribute("credentialserver",credentialserver,xrsl);
	{
		std::string n;
		get_attribute("notify",n,xrsl);
		notifications=notification(n);
	};
	/*client_software*/
	get_attribute("hostname",client_hostname,xrsl);
	get_attribute("lifetime",lifetime,xrsl);
	get_attribute("memory",memory,xrsl);
	get_attribute("count",count,xrsl);
	get_attribute("rerun",reruns,xrsl);
	get_attribute("cputime",cpu_time,xrsl);
    /*long wall_time;*/
	try {
		arguments.push_back((xrsl.GetRelation("executable")).GetSingleValue());
	} catch (XrslError) { arguments.push_back(""); };
	try {
		std::list<std::string> args =
		              (xrsl.GetRelation("arguments")).GetListValue();
		arguments.insert(arguments.end(),args.begin(),args.end());
	} catch (XrslError) { };
	try {
        std::list<XrslRelation> rels =
		            xrsl.GetAllRelations("runtimeenvironment");
		for(std::list<XrslRelation>::iterator i = rels.begin();
		                                        i!=rels.end();++i) {
			try {
				runtime_environments.push_back(
				            RuntimeEnvironment(i->GetSingleValue()));
			} catch (XrslError) { };
		};
	} catch (XrslError) { };
	try {
		std::list<XrslRelation> rels =
		            xrsl.GetAllRelations("middleware");
		for(std::list<XrslRelation>::iterator i = rels.begin();
		                                        i!=rels.end();++i) {
			try {
				middlewares.push_back(RuntimeEnvironment(i->GetSingleValue()));
			} catch (XrslError) { };
		};
	} catch (XrslError) { };
	try {
		std::list<XrslRelation> rels =
		            xrsl.GetAllRelations("jobreport");
		for(std::list<XrslRelation>::iterator i = rels.begin();
		                                        i!=rels.end();++i) {
			try {
				loggers.push_back(i->GetSingleValue());
			} catch (XrslError) { };
		};
	} catch (XrslError) { };
	try {
		std::list<std::list<std::string> > dfiles =
		         (xrsl.GetRelation("inputfiles")).GetDoubleListValue();
		for(std::list<std::list<std::string> >::iterator f = dfiles.begin();
		                                              f!=dfiles.end();++f) {
			std::list<std::string>::iterator i = f->begin();
			if(i == f->end()) continue;
			std::string name = *i; ++i;
			if(i != f->end()) {
				inputdata.push_back(InputFile(name,*i));
			} else {
				inputdata.push_back(InputFile(name,""));
			};
		};
	} catch (XrslError) { };
	try {
		std::list<std::list<std::string> > dfiles =
		         (xrsl.GetRelation("outputfiles")).GetDoubleListValue();
		for(std::list<std::list<std::string> >::iterator f = dfiles.begin();
		                                              f!=dfiles.end();++f) {
			std::list<std::string>::iterator i = f->begin();
			if(i == f->end()) continue;
			std::string name = *i; ++i;
			if(i != f->end()) {
				outputdata.push_back(OutputFile(name,*i));
			} else {
				outputdata.push_back(OutputFile(name,""));
			};
		};
	} catch (XrslError) { };
	try {
		executables=(xrsl.GetRelation("executables")).GetListValue();
	} catch (XrslError) { };

	try {
	    std::list<Xrsl> xrsls = xrsl.SplitOrRelation();
		if(xrsls.size() > 1) {
			for(std::list<Xrsl>::iterator i = xrsls.begin();
			                                i!=xrsls.end();++i) {
				JobRequest* jr = new JobRequest;
				((JobRequestXRSL*)jr)->set(*i);
				alternatives.push_back(jr);
			};
		};
	} catch (XrslError) { };

	return true;
}

bool JobRequestXRSL::print(std::string& s) throw(JobRequestError) {
	if(!xrsl_) return false;
	s=xrsl_->str();
	return true;
}

JobRequestXRSL::JobRequestXRSL(const JobRequest& j) throw(JobRequestError):xrsl_(NULL) {
  operator=(j);
}

JobRequestXRSL::JobRequestXRSL(const char* s) throw(JobRequestError):xrsl_(NULL) {
	if(!set(s)) throw JobRequestError("Can't parse xRSL");
}

JobRequestXRSL::JobRequestXRSL(const std::string& s) throw(JobRequestError):xrsl_(NULL) {
	if(!set(s.c_str())) throw JobRequestError("Can't parse xRSL");
}

JobRequestXRSL::JobRequestXRSL(std::istream& i) throw(JobRequestError):xrsl_(NULL) {
    char buf[256];
    std::string s;
    for(;!i.eof();) {
        i.get(buf,sizeof(buf),0);
        if(i.fail()) throw JobRequestError("Failed to read input stream of job request");
        if(buf[0] == 0) break;
        s.append(buf);
    };
    if(!set(s.c_str())) throw JobRequestError("Failed to parse job request");
}

JobRequestXRSL::~JobRequestXRSL(void) {
	if(xrsl_) delete xrsl_;
}


JobRequest& JobRequestXRSL::operator=(const JobRequest& j) throw (JobRequestError) {
	JobRequest::operator=(j);
	// Make xRSL
	if(xrsl_) delete xrsl_;
	xrsl_=new Xrsl;
	if(!xrsl_) return (*this);
	set_xrsl(*xrsl_);
	return (*this);
}

bool JobRequestXRSL::set_xrsl(Xrsl& xrsl) throw (JobRequestError) {
	try {
		add_attribute("jobname",job_name,xrsl);
		if(arguments.size() > 0) {
			xrsl.AddRelation(
			      XrslRelation("executable",operator_eq,*(arguments.begin())));
		};
		if(arguments.size() > 1) {
			std::list<std::string> args(++(arguments.begin()),arguments.end());
            xrsl.AddRelation(XrslRelation("arguments",operator_eq,args));
		};
		add_attribute("stdin", stdin, xrsl);
		add_attribute("stdout",stdout,xrsl);
		add_attribute("stderr",stderr,xrsl);
		add_attribute("notify",notification(notifications),xrsl);
		add_attribute("queue",queue,xrsl);
		add_attribute("acl",acl,xrsl);
		add_attribute("gmlog",gmlog,xrsl);
		add_attribute("credentialserver",credentialserver,xrsl);
		add_attribute("memory",memory,xrsl);
		add_attribute("cputime",cpu_time,xrsl);
        /*long wall_time;*/
		add_attribute("count",count,xrsl);
		add_attribute("rerun",reruns,xrsl);
		add_attribute("lifetime",lifetime,xrsl);
		for(std::list<RuntimeEnvironment>::const_iterator i =
		                 middlewares.begin();i!=middlewares.end();++i) {
			add_attribute("middleware",i->str(),xrsl);
		};
		add_attributes("jobreport",loggers,xrsl);
        add_attribute("executables",executables,xrsl);
        add_attribute("hostname",client_hostname,xrsl);
        /*std::string client_software;*/
		for(std::list<RuntimeEnvironment>::iterator i =
		      runtime_environments.begin();i!=runtime_environments.end();++i) {
			add_attribute("runtimeenvironment",i->str(),xrsl);
		};
		if(inputdata.size()) {
			std::list<std::list<std::string> > dfiles;
			for(std::list<InputFile>::iterator i = inputdata.begin();
			                                    i!=inputdata.end();++i) {
				std::list<std::string> dfile;
				dfile.push_back(i->name);
				dfile.push_back(i->source);
				dfiles.push_back(dfile);
			};
			xrsl.AddRelation(XrslRelation("inputfiles",operator_eq,dfiles));
		};
		if(outputdata.size()) {
			std::list<std::list<std::string> > dfiles;
			for(std::list<OutputFile>::iterator i = outputdata.begin();
			                                    i!=outputdata.end();++i) {
				std::list<std::string> dfile;
				dfile.push_back(i->name);
				dfile.push_back(i->destination);
				dfiles.push_back(dfile);
			};
			xrsl.AddRelation(XrslRelation("outputfiles",operator_eq,dfiles));
		};
		Xrsl xs(operator_or);
		int xs_n = 0;
		for(std::list<JobRequest*>::iterator i = alternatives.begin();
		                                     i!=alternatives.end();++i) {
			if(!(*i)) continue;
			Xrsl x;
			((JobRequestXRSL*)(*i))->set_xrsl(x);
			xs.AddXrsl(x); ++xs_n;
		};
		if(xs_n) xrsl.AddXrsl(xs);
	} catch (XrslError) { };
	return true;
}

