#include <arc/certificate.h>
#include <arc/common.h>
#include <arc/notify.h>
#include <arc/stringconv.h>
#include <arc/target.h>

#include <libintl.h>
#define _(A) dgettext("arclib", ((A)))


Target::Target(Queue aqueue, Xrsl axrsl) : Queue(aqueue) {
	xrsls = axrsl.SplitOrRelation();
}


void Target::AddXrsl(Xrsl axrsl) {
	xrsls = axrsl.SplitOrRelation();
}


std::list<Xrsl>& Target::GetXrsls() {
	return xrsls;
}


long Target::GetCputime(Xrsl xrsl) throw(XrslError) {

	bool cputimegiven = xrsl.IsRelation("cputime");
	bool walltimegiven = xrsl.IsRelation("walltime");
	bool gridtimegiven = xrsl.IsRelation("gridtime");
	bool benchmarksgiven = xrsl.IsRelation("benchmarks");

	if (cputimegiven && gridtimegiven)
		throw XrslError(_("Both cputime and gridtime given in the xrsl"));

	if (walltimegiven && gridtimegiven)
		throw XrslError(_("Both walltime and gridtime given in the xrsl"));

	if (cputimegiven && benchmarksgiven)
		throw XrslError(_("Both cputime and benchmarks given in the xrsl"));

	if (walltimegiven && benchmarksgiven)
		throw XrslError(_("Both walltime and benchmarks given in the xrsl"));

	if (benchmarksgiven && gridtimegiven)
		throw XrslError(_("Both gridtime and benchmarks given in the xrsl"));

	if (cputimegiven || walltimegiven) {
		long cputime = 0;
		long walltime = 0;
		if (cputimegiven) {
			std::string cpustr = xrsl.GetRelation("cputime").GetSingleValue();
			cputime = Seconds(cpustr);
		}
		if (walltimegiven) {
			std::string wastr = xrsl.GetRelation("walltime").GetSingleValue();
			walltime = Seconds(wastr);
		}

		if (walltime>=cputime) return walltime;
		if (walltime<cputime) return cputime;
	}
	
	if (gridtimegiven) {
		std::string gridstr = xrsl.GetRelation("gridtime").GetSingleValue();
		float cpufreq = cpu_freq;
		if (cpufreq == UNDEFINED) cpufreq = cluster.cpu_freq;
		if (cpufreq == UNDEFINED) {
			notify(WARNING) << _("Could not determine CPU frequency for queue")
			                << " " << name << "@" << cluster.hostname << " - "
			                << _("1 GHz is assumed") << std::endl;
			cpufreq = 1000;
		}

		return (long)(Seconds(gridstr)*2800.0/cpufreq);
	}

	if (benchmarksgiven) {
		long benchtime = UNDEFINED;

		std::map<std::string, float> queben = benchmarks;
		std::map<std::string, float> cluben = cluster.benchmarks;

		std::list<std::list<std::string> > bench =
		    xrsl.GetRelation("benchmarks").GetDoubleListValue();
		std::list<std::list<std::string> >::iterator it;

		for (it = bench.begin(); it != bench.end(); it++) {
			long thisbenchtime = UNDEFINED;

			std::list<std::string>::iterator listit = it->begin();
			std::string name = *listit;

			listit++;
			double value = stringtod(*listit);
			listit++;
			double time = Seconds(*listit);

			if (queben.find(name)!=queben.end()) {
				thisbenchtime = (long)(value*time/queben.find(name)->second);
			} else if (cluben.find(name)!=cluben.end()) {
				thisbenchtime = (long)(value*time/cluben.find(name)->second);
			}                     

			if (benchtime < thisbenchtime) benchtime = thisbenchtime;
		}

		if (benchtime!=UNDEFINED) return (long)benchtime;
	}

	return default_cpu_time;
}


std::list<Target> ConstructTargets(std::list<Queue> queues, Xrsl axrsl) {

	std::list<Target> targets;

	std::list<Queue>::iterator it;
	for (it = queues.begin(); it != queues.end(); it++) {

		/** Queue should be active. */
		if (it->status != "active") {
			notify(INFO) << _("Queue rejected because of status") << " "
			             << it->status << ": " << it->name << "@"
			             << it->cluster.hostname << std::endl;
			continue;
		}

		/** totalcpus should be > 0. */
		int totalcpus = 0;
		if (it->cluster.total_cpus != UNDEFINED)
			totalcpus = it->cluster.total_cpus;
		if (it->total_cpus != UNDEFINED)
			totalcpus = it->total_cpus;
		if (totalcpus == 0) {
			notify(INFO) << _("Queue rejected because it does not have any "
			                  "CPUs") << ": " << it->name << "@"
			             << it->cluster.hostname << std::endl;
			continue;
		}

		/** number of queued jobs should be less than maxqueuable. */
		if (it->max_queuable != UNDEFINED && it->queued > it->max_queuable) {
			notify(INFO) << _("Queue rejected because the number of queued "
			                  "jobs is larger than maximum allowed")
			             << ": " << it->name << "@"
			             << it->cluster.hostname << std::endl;
			continue;
		}

		/** the cluster's issuer-certificate must be installed locally. */
		std::list<Certificate> localcerts = GetCAList();
		std::list<Certificate>::iterator ceit;

		bool found = false;
		for (ceit = localcerts.begin(); ceit != localcerts.end(); ceit++) {
			if (ceit->GetSN()==it->cluster.issuer_ca) {
				found = true;
				break;
			}
		}

		if (found==false) {
			notify(INFO) << _("Queue rejected because the cluster's public "
			                  "keys are not installed on the client")
			             << ": " << it->name << "@"
			             << it->cluster.hostname << std::endl;
			continue;
		}

		/** The user should have free cpus. */
		if (it->users.begin()->free_cpus.empty()) {
			notify(INFO) << _("Queue rejected because the user is not "
			                  "authorized") << ": " << it->name << "@"
			             << it->cluster.hostname << std::endl;
			continue;
		}

		/** Max-running should be less than running. */
		if (it->max_running != UNDEFINED && it->max_running<=it->running) {
			notify(INFO) << _("Queue rejected because the number of running "
			                  "jobs is larger than the maximum allowed")
			             << ": " << it->name << "@"
			             << it->cluster.hostname << std::endl;
			continue;
		}

		Target target(*it, axrsl);
		targets.push_back(target);
	}

	return targets;
}
